#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from fairseq.models import (
    FairseqEncoder,
    FairseqEncoderModel,
    register_model,
    register_model_architecture,
)
from fairseq.modules.fairseq_dropout import FairseqDropout


default_conv_enc_config = """[
    (400, 13, 170, 0.2),
    (440, 14, 0, 0.214),
    (484, 15, 0, 0.22898),
    (532, 16, 0, 0.2450086),
    (584, 17, 0, 0.262159202),
    (642, 18, 0, 0.28051034614),
    (706, 19, 0, 0.30014607037),
    (776, 20, 0, 0.321156295296),
    (852, 21, 0, 0.343637235966),
    (936, 22, 0, 0.367691842484),
    (1028, 23, 0, 0.393430271458),
    (1130, 24, 0, 0.42097039046),
    (1242, 25, 0, 0.450438317792),
    (1366, 26, 0, 0.481969000038),
    (1502, 27, 0, 0.51570683004),
    (1652, 28, 0, 0.551806308143),
    (1816, 29, 0, 0.590432749713),
]"""


@register_model("asr_w2l_conv_glu_encoder")
class W2lConvGluEncoderModel(FairseqEncoderModel):
    def __init__(self, encoder):
        super().__init__(encoder)

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        parser.add_argument(
            "--input-feat-per-channel",
            type=int,
            metavar="N",
            help="encoder input dimension per input channel",
        )
        parser.add_argument(
            "--in-channels",
            type=int,
            metavar="N",
            help="number of encoder input channels",
        )
        parser.add_argument(
            "--conv-enc-config",
            type=str,
            metavar="EXPR",
            help="""
    an array of tuples each containing the configuration of one conv layer
    [(out_channels, kernel_size, padding, dropout), ...]
            """,
        )

    @classmethod
    def build_model(cls, args, task):
        """Build a new model instance."""
        conv_enc_config = getattr(args, "conv_enc_config", default_conv_enc_config)
        encoder = W2lConvGluEncoder(
            vocab_size=len(task.target_dictionary),
            input_feat_per_channel=args.input_feat_per_channel,
            in_channels=args.in_channels,
            conv_enc_config=eval(conv_enc_config),
        )
        return cls(encoder)

    def get_normalized_probs(self, net_output, log_probs, sample=None):
        lprobs = super().get_normalized_probs(net_output, log_probs, sample)
        lprobs.batch_first = False
        return lprobs


class W2lConvGluEncoder(FairseqEncoder):
    def __init__(
        self, vocab_size, input_feat_per_channel, in_channels, conv_enc_config
    ):
        super().__init__(None)

        self.input_dim = input_feat_per_channel
        if in_channels != 1:
            raise ValueError("only 1 input channel is currently supported")

        self.conv_layers = nn.ModuleList()
        self.linear_layers = nn.ModuleList()
        self.dropouts = []
        cur_channels = input_feat_per_channel

        for out_channels, kernel_size, padding, dropout in conv_enc_config:
            layer = nn.Conv1d(cur_channels, out_channels, kernel_size, padding=padding)
            layer.weight.data.mul_(math.sqrt(3))  # match wav2letter init
            self.conv_layers.append(nn.utils.weight_norm(layer))
            self.dropouts.append(
                FairseqDropout(dropout, module_name=self.__class__.__name__)
            )
            if out_channels % 2 != 0:
                raise ValueError("odd # of out_channels is incompatible with GLU")
            cur_channels = out_channels // 2  # halved by GLU

        for out_channels in [2 * cur_channels, vocab_size]:
            layer = nn.Linear(cur_channels, out_channels)
            layer.weight.data.mul_(math.sqrt(3))
            self.linear_layers.append(nn.utils.weight_norm(layer))
            cur_channels = out_channels // 2

    def forward(self, src_tokens, src_lengths, **kwargs):

        """
        src_tokens: padded tensor (B, T, C * feat)
        src_lengths: tensor of original lengths of input utterances (B,)
        """
        B, T, _ = src_tokens.size()
        x = src_tokens.transpose(1, 2).contiguous()  # (B, feat, T) assuming C == 1

        for layer_idx in range(len(self.conv_layers)):
            x = self.conv_layers[layer_idx](x)
            x = F.glu(x, dim=1)
            x = self.dropouts[layer_idx](x)

        x = x.transpose(1, 2).contiguous()  # (B, T, 908)
        x = self.linear_layers[0](x)
        x = F.glu(x, dim=2)
        x = self.dropouts[-1](x)
        x = self.linear_layers[1](x)

        assert x.size(0) == B
        assert x.size(1) == T

        encoder_out = x.transpose(0, 1)  # (T, B, vocab_size)

        # need to debug this -- find a simpler/elegant way in pytorch APIs
        encoder_padding_mask = (
            torch.arange(T).view(1, T).expand(B, -1).to(x.device)
            >= src_lengths.view(B, 1).expand(-1, T)
        ).t()  # (B x T) -> (T x B)

        return {
            "encoder_out": encoder_out,  # (T, B, vocab_size)
            "encoder_padding_mask": encoder_padding_mask,  # (T, B)
        }

    def reorder_encoder_out(self, encoder_out, new_order):
        encoder_out["encoder_out"] = encoder_out["encoder_out"].index_select(
            1, new_order
        )
        encoder_out["encoder_padding_mask"] = encoder_out[
            "encoder_padding_mask"
        ].index_select(1, new_order)
        return encoder_out

    def max_positions(self):
        """Maximum input length supported by the encoder."""
        return (1e6, 1e6)  # an arbitrary large number


@register_model_architecture("asr_w2l_conv_glu_encoder", "w2l_conv_glu_enc")
def w2l_conv_glu_enc(args):
    args.input_feat_per_channel = getattr(args, "input_feat_per_channel", 80)
    args.in_channels = getattr(args, "in_channels", 1)
    args.conv_enc_config = getattr(args, "conv_enc_config", default_conv_enc_config)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Replabel transforms for use with flashlight's ASG criterion.
"""


def replabel_symbol(i):
    """
    Replabel symbols used in flashlight, currently just "1", "2", ...
    This prevents training with numeral tokens, so this might change in the future
    """
    return str(i)


def pack_replabels(tokens, dictionary, max_reps):
    """
    Pack a token sequence so that repeated symbols are replaced by replabels
    """
    if len(tokens) == 0 or max_reps <= 0:
        return tokens

    replabel_value_to_idx = [0] * (max_reps + 1)
    for i in range(1, max_reps + 1):
        replabel_value_to_idx[i] = dictionary.index(replabel_symbol(i))

    result = []
    prev_token = -1
    num_reps = 0
    for token in tokens:
        if token == prev_token and num_reps < max_reps:
            num_reps += 1
        else:
            if num_reps > 0:
                result.append(replabel_value_to_idx[num_reps])
                num_reps = 0
            result.append(token)
            prev_token = token
    if num_reps > 0:
        result.append(replabel_value_to_idx[num_reps])
    return result


def unpack_replabels(tokens, dictionary, max_reps):
    """
    Unpack a token sequence so that replabels are replaced by repeated symbols
    """
    if len(tokens) == 0 or max_reps <= 0:
        return tokens

    replabel_idx_to_value = {}
    for i in range(1, max_reps + 1):
        replabel_idx_to_value[dictionary.index(replabel_symbol(i))] = i

    result = []
    prev_token = -1
    for token in tokens:
        try:
            for _ in range(replabel_idx_to_value[token]):
                result.append(prev_token)
            prev_token = -1
        except KeyError:
            result.append(token)
            prev_token = token
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Utils functions
"""

import json
import logging
import os
import tarfile
from pathlib import Path
from urllib.request import urlopen
from io import BytesIO
from zipfile import ZipFile

import requests
from git import Repo
from tqdm import tqdm

from wow_ai_mms import constants

logger = logging.getLogger(__name__)


def download_file(url: str, download_dir=None, chunk_size=1024) -> str:
    """
    Helper function to download models and other required files
    :param url: URL of the file
    :param download_dir: Where to store the file
    :param chunk_size: size of the download chunk

    :return: Absolute path of the downloaded model
    """

    os.makedirs(download_dir, exist_ok=True)
    file_name = os.path.basename(url)
    file_path = Path(download_dir) / file_name
    # check if the file is already there
    if file_path.exists():
        logging.info(f"File '{file_name}' already exists in {download_dir}")
    else:
        # download it
        resp = requests.get(url, stream=True)
        total = int(resp.headers.get('content-length', 0))

        progress_bar = tqdm(desc=f"Downloading File {file_name} ...",
                            total=total,
                            unit='iB',
                            unit_scale=True,
                            unit_divisor=1024)

        try:
            with open(file_path, 'wb') as file, progress_bar:
                for data in resp.iter_content(chunk_size=chunk_size):
                    size = file.write(data)
                    progress_bar.update(size)
            logging.info(f"Model downloaded to {file_path.absolute()}")
        except Exception as e:
            # error download, just remove the file
            os.remove(file_path)
            raise e
    return str(file_path.absolute())


def download_and_extract(url, extract_to='.'):
    """
    Downloads and unzips a zip folder
    Will be used to download uroman source code
    :param url: the file URL
    :param extract_to: extract path
    :return: None
    """
    logger.info(f"Downloading file '{url}' and extracting to '{extract_to}' ...")
    http_response = urlopen(url)
    file_name = os.path.basename(url)
    if file_name.endswith('.zip'):
        zipfile = ZipFile(BytesIO(http_response.read()))
        zipfile.extractall(path=extract_to)
    else:
        tar = tarfile.open(BytesIO(http_response.read()), "r:gz")
        tar.extractall(path=extract_to)
        tar.close()


def clone(repo_url, destination_folder: Path):
    if destination_folder.exists():
        logger.info(f"{destination_folder} already exists")
    else:
        logger.info(f"Cloning from {repo_url} tp {destination_folder} ...")
        Repo.clone_from(repo_url, str(destination_folder.resolve()))
    return destination_folder
def get_uroman():
    clone(constants.UROMAN_URL, constants.UROMAN_DIR)
    return constants.UROMAN_DIR / 'bin'

def get_lang_info(lang: str) -> dict:
    """
    Returns more info about a language,
    :param lang: the ISO 693-3 language code
    :return: dict of info
    """
    with open(constants.MMS_LANGS_FILE) as f:
        data = json.load(f)
        return data[lang]


def get_transcript_segments(transcript: str, t_type: str = 'segment', max_segment_len: int = 24):
    """
    A helper function to fragment the transcript to segments
    <Quick implementation, Not perfect (barely works), needs improvements>

    :param transcript: results of the ASR model
    :param t_type: one of [`word`, `segment`, `char`]
    :param max_segment_len: the maximum length of the segment
    :return: list of segments
    """
    res = []
    if t_type == 'word':
        res = transcript.split()
    elif t_type == 'char':
        s = ''
        for c in transcript:
            if c == ' ':
                continue
            if len(s) >= max_segment_len:
                res.append(s)
                s = c
            else:
                s += c
        res.append(s)
    else:
        s = ''
        words = transcript.strip().split()
        for word in words:
            new_word = s + ' ' + word
            if len(new_word) > max_segment_len:
                res.append(s.strip())
                s = word
            else:
                s = new_word
        res.append(s)

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file contains a class definition to use the TTS models from [Meta's Massively Multilingual Speech (MMS) project](https://github.com/facebookresearch/fairseq/tree/main/examples/mms)
"""

__author__ = "Abdeladim S."
__copyright__ = "Copyright 2023,"

import json
import os
import sys
import tarfile
import logging
from pathlib import Path
from typing import List, Tuple
import torch
import soundfile as sf
import site



import wow_ai_mms.utils as wow_ai_mms_utils
from wow_ai_mms._logger import set_log_level
from wow_ai_mms import constants

logger = logging.getLogger(__name__)


class TTSModel:
    """
    TTS class model

    Example usage:

    ```python
    from wow_ai_mms.models.tts import TTSModel

    tts = TTSModel('eng')
    res = tts.synthesize("This is a simple example")
    tts.save(res)
    ```
    """

    def __init__(self,
                 lang: str,
                 model_dir: str = None,
                 log_level: int = logging.INFO):
        """
        Use a TTS model by its language ISO ID.
        The model will be downloaded automatically

        :param lang: TTS model language
        :param log_level: log level
        """
        self.log_level = log_level
        self.lang = lang
        set_log_level(log_level)
        # check if models_dir is provided
        if model_dir is not None:
            # verify if all files exist
            model_dir_path = Path(model_dir)
        else:
            model_dir_path = self._download_tts_model_files(lang)

        self.cp = model_dir_path / "G_100000.pth"
        assert self.cp.exists(), f"G_100000.pth not found in {str(model_dir_path)}"
        self.config = model_dir_path / "config.json"
        assert self.config.exists(), f"config.json not found in {str(model_dir_path)}"
        self.vocab = model_dir_path / "vocab.txt"
        assert self.vocab.exists(), f"vocab.txt not found in {str(model_dir_path)}"
        self.uroman_dir_path = None

        self._setup()

    def _setup(self) -> None:
        """
        Helper function to setup different required packages
        :return: None
        """
        # clone Uroman
        wow_ai_mms_utils.clone(constants.UROMAN_URL, constants.UROMAN_DIR)
        # clone VITS
        wow_ai_mms_utils.clone(constants.VITS_URL, constants.VITS_DIR)
        sys.path.append(str(constants.VITS_DIR.resolve()))
        # clone Fairseq
        wow_ai_mms_utils.clone(constants.FAIRSEQ_URL, constants.FAIRSEQ_DIR)
        sys.path.append(str(constants.FAIRSEQ_DIR.resolve()))

        # build monotonic ext
        monotonic_ext_dir = constants.VITS_DIR / 'monotonic_align' / 'monotonic_align'
        Path.mkdir(monotonic_ext_dir, exist_ok=True)
        cwd = os.getcwd()
        if len(list(monotonic_ext_dir.iterdir())) > 0:
            # extension is probably already there
            return
        else:
            from distutils.core import run_setup
            os.chdir(str(constants.VITS_DIR / 'monotonic_align'))
            run_setup(str(constants.VITS_DIR / 'monotonic_align' / 'setup.py'), script_args=['build_ext', '--inplace'],
                      stop_after='run')
            # os.chdir(cwd)
    @staticmethod
    def _download_tts_model_files(lang: str):
        """
        Downloads and extracts model files into the TTS directory (::: constants.TTS_DIR)
        :param lang:
        :return:
        """
        model_dir = constants.TTS_DIR / lang
        url = constants.TTS_MODELS_BASE_URL + f"{lang}.tar.gz"
        compressed_file = wow_ai_mms_utils.download_file(url, str(constants.TTS_DIR))
        # extract it
        logger.info(f"Extracting {compressed_file} to {constants.TTS_DIR}")
        tar = tarfile.open(compressed_file, "r:gz")
        tar.extractall(path=constants.TTS_DIR)
        tar.close()
        return model_dir

    @staticmethod
    def get_supported_langs() -> List[str]:
        """
        Helper function to get supported ISO 693-3 languages by the TTS models
        Source <https://dl.fbaipublicfiles.com/mms/misc/language_coverage_mms.html>
        :return: list of supported languages
        """
        with open(constants.MMS_LANGS_FILE) as f:
            data = json.load(f)
            return [key for key in data if data[key]['TTS']]

    def synthesize(self, txt: str, device=None):
        """
         Synthesizes the text provided as input.

        :param txt: Text
        :param lang: Language
        :param device: Pytorch device (cpu/cuda)
        :return: Tuple(data, sample_rate)
        """
        cwd = os.getcwd()
        os.chdir(constants.VITS_DIR)
        from utils import get_hparams_from_file, load_checkpoint
        from models import SynthesizerTrn
        os.chdir(constants.FAIRSEQ_DIR)
        from examples.mms.tts.infer import TextMapper
        os.chdir(cwd)
        if device is None:
            if torch.cuda.is_available():
                device = 'cuda'
            else:
                device = 'cpu'
        else:
            assert device in ['cpu', 'cuda']

        text_mapper = TextMapper(str(self.vocab))
        hps = get_hparams_from_file(str(self.config))
        net_g = SynthesizerTrn(
            len(text_mapper.symbols),
            hps.data.filter_length // 2 + 1,
            hps.train.segment_size // hps.data.hop_length,
            **hps.model)
        net_g.to(device)
        _ = net_g.eval()
        g_pth = self.cp
        logger.info(f"loading {g_pth} ...")
        _ = load_checkpoint(g_pth, net_g, None)
        logger.info(f"text: {txt}")
        is_uroman = hps.data.training_files.split('.')[-1] == 'uroman'
        if is_uroman:
            uroman_pl = str(self.uroman_dir_path / "uroman.pl")
            txt = text_mapper.uromanize(txt, uroman_pl)
            logger.info(f"uroman text: {txt}")
        txt = txt.lower()
        txt = text_mapper.filter_oov(txt, lang=self.lang)
        stn_tst = text_mapper.get_text(txt, hps)
        # inference
        with torch.no_grad():
            x_tst = stn_tst.unsqueeze(0).to(device)
            x_tst_lengths = torch.LongTensor([stn_tst.size(0)]).to(device)
            hyp = net_g.infer(
                x_tst, x_tst_lengths, noise_scale=.667,
                noise_scale_w=0.8, length_scale=1.0
            )[0][0, 0].cpu().float().numpy()

        return hyp, hps.data.sampling_rate

    def save(self, tts_data: Tuple, out_file='out.wav') -> Path:
        """
        Saves the results of the `synthesize` function to a file

        :param tts_data: tts_data: a tuple of `wav data array` and `sample rate`
        :param out_file: output file path

        :return: out_file absolute path
        """
        set_log_level(self.log_level)
        logger.info(f"Saving audio file to {out_file}")
        sf.write(out_file, tts_data[0], tts_data[1])
        return out_file



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Constants
"""
import logging
from pathlib import Path

import numpy as np
from platformdirs import *

PACKAGE_NAME = 'wow_ai_mms'
LOGGING_LEVEL = logging.INFO

PACKAGE_DATA_DIR = user_data_dir(PACKAGE_NAME)


TTS_DIR = (Path(PACKAGE_DATA_DIR) / 'tts').resolve()
TTS_MODELS_BASE_URL = "https://dl.fbaipublicfiles.com/mms/tts/"  # lang.tar.gz
VITS_URL = "https://github.com/jaywalnut310/vits"
VITS_DIR = TTS_DIR / 'vits'

FAIRSEQ_URL = "https://github.com/facebookresearch/fairseq"
FAIRSEQ_DIR = Path(PACKAGE_DATA_DIR) / 'fairseq'

CFG = {
  '_name': None,
  'task': {
    '_name': 'audio_finetuning',
    'data': '',
    'labels': 'ltr'
  },
  'decoding': {
    '_name': None,
    'nbest': 1,
    'unitlm': False,
    'lmpath': '???',
    'lexicon': None,
    'beam': 50,
    'beamthreshold': 50.0,
    'beamsizetoken': None,
    'wordscore': -1.0,
    'unkweight': -np.inf,
    'silweight': 0.0,
    'lmweight': 2.0,
    'type': 'viterbi',
    'unique_wer_file': False,
    'results_path': ''
  },
  'common': {
    '_name': None,
    'no_progress_bar': False,
    'log_interval': 100,
    'log_format': None,
    'log_file': None,
    'aim_repo': None,
    'aim_run_hash': None,
    'tensorboard_logdir': None,
    'wandb_project': None,
    'azureml_logging': False,
    'seed': 1,
    'cpu': False,
    'tpu': False,
    'bf16': False,
    'memory_efficient_bf16': False,
    'fp16': False,
    'memory_efficient_fp16': False,
    'fp16_no_flatten_grads': False,
    'fp16_init_scale': 128,
    'fp16_scale_window': None,
    'fp16_scale_tolerance': 0.0,
    'on_cpu_convert_precision': False,
    'min_loss_scale': 0.0001,
    'threshold_loss_scale': None,
    'amp': False,
    'amp_batch_retries': 2,
    'amp_init_scale': 128,
    'amp_scale_window': None,
    'user_dir': None,
    'empty_cache_freq': 0,
    'all_gather_list_size': 16384,
    'model_parallel_size': 1,
    'quantization_config_path': None,
    'profile': False,
    'reset_logging': False,
    'suppress_crashes': False,
    'use_plasma_view': False,
    'plasma_path': '/tmp/plasma'
  },
  'common_eval': {
    '_name': None,
    'path': '',
    'post_process': 'letter',
    'quiet': False,
    'model_overrides': '{}',
    'results_path': None
  },
  'checkpoint': {
    '_name': None,
    'save_dir': 'checkpoints',
    'restore_file': 'checkpoint_last.pt',
    'continue_once': None,
    'finetune_from_model': None,
    'reset_dataloader': False,
    'reset_lr_scheduler': False,
    'reset_meters': False,
    'reset_optimizer': False,
    'optimizer_overrides': '{}',
    'save_interval': 1,
    'save_interval_updates': 0,
    'keep_interval_updates': -1,
    'keep_interval_updates_pattern': -1,
    'keep_last_epochs': -1,
    'keep_best_checkpoints': -1,
    'no_save': False,
    'no_epoch_checkpoints': False,
    'no_last_checkpoints': False,
    'no_save_optimizer_state': False,
    'best_checkpoint_metric': 'loss',
    'maximize_best_checkpoint_metric': False,
    'patience': -1,
    'checkpoint_suffix': '',
    'checkpoint_shard_count': 1,
    'load_checkpoint_on_all_dp_ranks': False,
    'write_checkpoints_asynchronously': False,
    'model_parallel_size': 1
  },
  'distributed_training': {
    '_name': None,
    'distributed_world_size': 1,
    'distributed_num_procs': 1,
    'distributed_rank': 0,
    'distributed_backend': 'nccl',
    'distributed_init_method': None,
    'distributed_port': -1,
    'device_id': 0,
    'distributed_no_spawn': False,
    'ddp_backend': 'legacy_ddp',
    'ddp_comm_hook': 'none',
    'bucket_cap_mb': 25,
    'fix_batches_to_gpus': False,
    'find_unused_parameters': False,
    'gradient_as_bucket_view': False,
    'fast_stat_sync': False,
    'heartbeat_timeout': -1,
    'broadcast_buffers': False,
    'slowmo_momentum': None,
    'slowmo_base_algorithm': 'localsgd',
    'localsgd_frequency': 3,
    'nprocs_per_node': 1,
    'pipeline_model_parallel': False,
    'pipeline_balance': None,
    'pipeline_devices': None,
    'pipeline_chunks': 0,
    'pipeline_encoder_balance': None,
    'pipeline_encoder_devices': None,
    'pipeline_decoder_balance': None,
    'pipeline_decoder_devices': None,
    'pipeline_checkpoint': 'never',
    'zero_sharding': 'none',
    'fp16': False,
    'memory_efficient_fp16': True,
    'tpu': False,
    'no_reshard_after_forward': False,
    'fp32_reduce_scatter': False,
    'cpu_offload': False,
    'use_sharded_state': False,
    'not_fsdp_flatten_parameters': False
  },
  'dataset': {
    '_name': None,
    'num_workers': 1,
    'skip_invalid_size_inputs_valid_test': False,
    'max_tokens': 4000000,
    'batch_size': None,
    'required_batch_size_multiple': 1,
    'required_seq_len_multiple': 1,
    'dataset_impl': None,
    'data_buffer_size': 10,
    'train_subset': 'train',
    'valid_subset': 'valid',
    'combine_valid_subsets': None,
    'ignore_unused_valid_subsets': False,
    'validate_interval': 1,
    'validate_interval_updates': 0,
    'validate_after_updates': 0,
    'fixed_validation_seed': None,
    'disable_validation': False,
    'max_tokens_valid': 4000000,
    'batch_size_valid': None,
    'max_valid_steps': None,
    'curriculum': 0,
    'gen_subset': 'eng:dev',
    'num_shards': 1,
    'shard_id': 0,
    'grouped_shuffling': False,
    'update_epoch_batch_itr': False,
    'update_ordered_indices_seed': False
  },
  'is_ax': False
}

HYPO_WORDS_FILE = 'hypo.word'

ALIGNMENT_MODEL_URL = "https://dl.fbaipublicfiles.com/mms/torchaudio/ctc_alignment_mling_uroman/model.pt"
ALIGNMENT_DICTIONARY_URL = "https://dl.fbaipublicfiles.com/mms/torchaudio/ctc_alignment_mling_uroman/dictionary.txt"

UROMAN_URL = "https://github.com/isi-nlp/uroman"
UROMAN_DIR = Path(PACKAGE_DATA_DIR) / 'uroman'

MMS_LANGS_FILE = (Path(__file__).parent / 'data' / 'mms_langs.json').resolve()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import re
from collections import deque
from enum import Enum

import numpy as np


"""
    Utility modules for computation of Word Error Rate,
    Alignments, as well as more granular metrics like
    deletion, insersion and substitutions.
"""


class Code(Enum):
    match = 1
    substitution = 2
    insertion = 3
    deletion = 4


class Token(object):
    def __init__(self, lbl="", st=np.nan, en=np.nan):
        if np.isnan(st):
            self.label, self.start, self.end = "", 0.0, 0.0
        else:
            self.label, self.start, self.end = lbl, st, en


class AlignmentResult(object):
    def __init__(self, refs, hyps, codes, score):
        self.refs = refs  # std::deque<int>
        self.hyps = hyps  # std::deque<int>
        self.codes = codes  # std::deque<Code>
        self.score = score  # float


def coordinate_to_offset(row, col, ncols):
    return int(row * ncols + col)


def offset_to_row(offset, ncols):
    return int(offset / ncols)


def offset_to_col(offset, ncols):
    return int(offset % ncols)


def trimWhitespace(str):
    return re.sub(" +", " ", re.sub(" *$", "", re.sub("^ *", "", str)))


def str2toks(str):
    pieces = trimWhitespace(str).split(" ")
    toks = []
    for p in pieces:
        toks.append(Token(p, 0.0, 0.0))
    return toks


class EditDistance(object):
    def __init__(self, time_mediated):
        self.time_mediated_ = time_mediated
        self.scores_ = np.nan  # Eigen::Matrix<float, Eigen::Dynamic, Eigen::Dynamic>
        self.backtraces_ = (
            np.nan
        )  # Eigen::Matrix<size_t, Eigen::Dynamic, Eigen::Dynamic> backtraces_;
        self.confusion_pairs_ = {}

    def cost(self, ref, hyp, code):
        if self.time_mediated_:
            if code == Code.match:
                return abs(ref.start - hyp.start) + abs(ref.end - hyp.end)
            elif code == Code.insertion:
                return hyp.end - hyp.start
            elif code == Code.deletion:
                return ref.end - ref.start
            else:  # substitution
                return abs(ref.start - hyp.start) + abs(ref.end - hyp.end) + 0.1
        else:
            if code == Code.match:
                return 0
            elif code == Code.insertion or code == Code.deletion:
                return 3
            else:  # substitution
                return 4

    def get_result(self, refs, hyps):
        res = AlignmentResult(refs=deque(), hyps=deque(), codes=deque(), score=np.nan)

        num_rows, num_cols = self.scores_.shape
        res.score = self.scores_[num_rows - 1, num_cols - 1]

        curr_offset = coordinate_to_offset(num_rows - 1, num_cols - 1, num_cols)

        while curr_offset != 0:
            curr_row = offset_to_row(curr_offset, num_cols)
            curr_col = offset_to_col(curr_offset, num_cols)

            prev_offset = self.backtraces_[curr_row, curr_col]

            prev_row = offset_to_row(prev_offset, num_cols)
            prev_col = offset_to_col(prev_offset, num_cols)

            res.refs.appendleft(curr_row - 1)  # Note: this was .push_front() in C++
            res.hyps.appendleft(curr_col - 1)
            if curr_row - 1 == prev_row and curr_col == prev_col:
                res.codes.appendleft(Code.deletion)
            elif curr_row == prev_row and curr_col - 1 == prev_col:
                res.codes.appendleft(Code.insertion)
            else:
                # assert(curr_row - 1 == prev_row and curr_col - 1 == prev_col)
                ref_str = refs[res.refs[0]].label
                hyp_str = hyps[res.hyps[0]].label

                if ref_str == hyp_str:
                    res.codes.appendleft(Code.match)
                else:
                    res.codes.appendleft(Code.substitution)

                    confusion_pair = "%s -> %s" % (ref_str, hyp_str)
                    if confusion_pair not in self.confusion_pairs_:
                        self.confusion_pairs_[confusion_pair] = 1
                    else:
                        self.confusion_pairs_[confusion_pair] += 1

            curr_offset = prev_offset

        return res

    def align(self, refs, hyps):
        if len(refs) == 0 and len(hyps) == 0:
            return np.nan

        # NOTE: we're not resetting the values in these matrices because every value
        # will be overridden in the loop below. If this assumption doesn't hold,
        # be sure to set all entries in self.scores_ and self.backtraces_ to 0.
        self.scores_ = np.zeros((len(refs) + 1, len(hyps) + 1))
        self.backtraces_ = np.zeros((len(refs) + 1, len(hyps) + 1))

        num_rows, num_cols = self.scores_.shape

        for i in range(num_rows):
            for j in range(num_cols):
                if i == 0 and j == 0:
                    self.scores_[i, j] = 0.0
                    self.backtraces_[i, j] = 0
                    continue

                if i == 0:
                    self.scores_[i, j] = self.scores_[i, j - 1] + self.cost(
                        None, hyps[j - 1], Code.insertion
                    )
                    self.backtraces_[i, j] = coordinate_to_offset(i, j - 1, num_cols)
                    continue

                if j == 0:
                    self.scores_[i, j] = self.scores_[i - 1, j] + self.cost(
                        refs[i - 1], None, Code.deletion
                    )
                    self.backtraces_[i, j] = coordinate_to_offset(i - 1, j, num_cols)
                    continue

                # Below here both i and j are greater than 0
                ref = refs[i - 1]
                hyp = hyps[j - 1]
                best_score = self.scores_[i - 1, j - 1] + (
                    self.cost(ref, hyp, Code.match)
                    if (ref.label == hyp.label)
                    else self.cost(ref, hyp, Code.substitution)
                )

                prev_row = i - 1
                prev_col = j - 1
                ins = self.scores_[i, j - 1] + self.cost(None, hyp, Code.insertion)
                if ins < best_score:
                    best_score = ins
                    prev_row = i
                    prev_col = j - 1

                delt = self.scores_[i - 1, j] + self.cost(ref, None, Code.deletion)
                if delt < best_score:
                    best_score = delt
                    prev_row = i - 1
                    prev_col = j

                self.scores_[i, j] = best_score
                self.backtraces_[i, j] = coordinate_to_offset(
                    prev_row, prev_col, num_cols
                )

        return self.get_result(refs, hyps)


class WERTransformer(object):
    def __init__(self, hyp_str, ref_str, verbose=True):
        self.ed_ = EditDistance(False)
        self.id2oracle_errs_ = {}
        self.utts_ = 0
        self.words_ = 0
        self.insertions_ = 0
        self.deletions_ = 0
        self.substitutions_ = 0

        self.process(["dummy_str", hyp_str, ref_str])

        if verbose:
            print("'%s' vs '%s'" % (hyp_str, ref_str))
            self.report_result()

    def process(self, input):  # std::vector<std::string>&& input
        if len(input) < 3:
            print(
                "Input must be of the form <id> ... <hypo> <ref> , got ",
                len(input),
                " inputs:",
            )
            return None

        # Align
        # std::vector<Token> hyps;
        # std::vector<Token> refs;

        hyps = str2toks(input[-2])
        refs = str2toks(input[-1])

        alignment = self.ed_.align(refs, hyps)
        if alignment is None:
            print("Alignment is null")
            return np.nan

        # Tally errors
        ins = 0
        dels = 0
        subs = 0
        for code in alignment.codes:
            if code == Code.substitution:
                subs += 1
            elif code == Code.insertion:
                ins += 1
            elif code == Code.deletion:
                dels += 1

        # Output
        row = input
        row.append(str(len(refs)))
        row.append(str(ins))
        row.append(str(dels))
        row.append(str(subs))
        # print(row)

        # Accumulate
        kIdIndex = 0
        kNBestSep = "/"

        pieces = input[kIdIndex].split(kNBestSep)

        if len(pieces) == 0:
            print(
                "Error splitting ",
                input[kIdIndex],
                " on '",
                kNBestSep,
                "', got empty list",
            )
            return np.nan

        id = pieces[0]
        if id not in self.id2oracle_errs_:
            self.utts_ += 1
            self.words_ += len(refs)
            self.insertions_ += ins
            self.deletions_ += dels
            self.substitutions_ += subs
            self.id2oracle_errs_[id] = [ins, dels, subs]
        else:
            curr_err = ins + dels + subs
            prev_err = np.sum(self.id2oracle_errs_[id])
            if curr_err < prev_err:
                self.id2oracle_errs_[id] = [ins, dels, subs]

        return 0

    def report_result(self):
        # print("----------  Summary ---------------")
        if self.words_ == 0:
            print("No words counted")
            return

        # 1-best
        best_wer = (
            100.0
            * (self.insertions_ + self.deletions_ + self.substitutions_)
            / self.words_
        )

        print(
            "\tWER = %0.2f%% (%i utts, %i words, %0.2f%% ins, "
            "%0.2f%% dels, %0.2f%% subs)"
            % (
                best_wer,
                self.utts_,
                self.words_,
                100.0 * self.insertions_ / self.words_,
                100.0 * self.deletions_ / self.words_,
                100.0 * self.substitutions_ / self.words_,
            )
        )

    def wer(self):
        if self.words_ == 0:
            wer = np.nan
        else:
            wer = (
                100.0
                * (self.insertions_ + self.deletions_ + self.substitutions_)
                / self.words_
            )
        return wer

    def stats(self):
        if self.words_ == 0:
            stats = {}
        else:
            wer = (
                100.0
                * (self.insertions_ + self.deletions_ + self.substitutions_)
                / self.words_
            )
            stats = dict(
                {
                    "wer": wer,
                    "utts": self.utts_,
                    "numwords": self.words_,
                    "ins": self.insertions_,
                    "dels": self.deletions_,
                    "subs": self.substitutions_,
                    "confusion_pairs": self.ed_.confusion_pairs_,
                }
            )
        return stats


def calc_wer(hyp_str, ref_str):
    t = WERTransformer(hyp_str, ref_str, verbose=0)
    return t.wer()


def calc_wer_stats(hyp_str, ref_str):
    t = WERTransformer(hyp_str, ref_str, verbose=0)
    return t.stats()


def get_wer_alignment_codes(hyp_str, ref_str):
    """
    INPUT: hypothesis string, reference string
    OUTPUT: List of alignment codes (intermediate results from WER computation)
    """
    t = WERTransformer(hyp_str, ref_str, verbose=0)
    return t.ed_.align(str2toks(ref_str), str2toks(hyp_str)).codes


def merge_counts(x, y):
    # Merge two hashes which have 'counts' as their values
    # This can be used for example to merge confusion pair counts
    #   conf_pairs = merge_counts(conf_pairs, stats['confusion_pairs'])
    for k, v in y.items():
        if k not in x:
            x[k] = 0
        x[k] += v
    return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file contains a class definition to use the ASR models from [Meta's Massively Multilingual Speech (MMS) project](https://github.com/facebookresearch/fairseq/tree/main/examples/mms)
"""

__author__ = "Tony Shark"
__copyright__ = "Copyright 2023,"


import os
import atexit
import json
import re
import sys
import tempfile
import logging
from pathlib import Path
from typing import Union, List
import torch
from omegaconf import OmegaConf
from pydub import AudioSegment
from pydub.utils import mediainfo

# fix importing from fairseq.examples
# import site
# sys.path.append(str(Path(site.getsitepackages()[0]) / 'fairseq'))
# try:
#     from fairseq.examples.speech_recognition.new.infer import hydra_main
# except ImportError:
#     from examples.speech_recognition.new.infer import hydra_main


from wow_ai_mms import utils as wow_ai_mms_utils
from wow_ai_mms._logger import set_log_level
from wow_ai_mms.models.alignment import AlignmentModel
from wow_ai_mms import constants

logger = logging.getLogger(__name__)


class ASRModel:
    """
    MMS ASR class model

    Example usage:

    ```python
    from wow_ai_mms.models.asr import ASRModel

    asr = ASRModel(model='/path/to/mms/model')
    files = ['path/to/media_file_1', 'path/to/media_file_2']
    transcriptions = asr.transcribe(files, lang='eng', align=False)
    for i, transcription in enumerate(transcriptions):
        print(f">>> file {files[i]}")
        print(transcription)
    
    for i, transcription in enumerate(transcriptions):
        # print(f">>> file {files[i]}")
        for segment in transcription:
            print(f"{segment['start_time']} -> {segment['end_time']}: {segment['text']}")
        print("----")
    ```
    """

    def __init__(self,
                 model: str,
                 log_level: int = logging.INFO):
        """
        :param model: path to the asr model <https://github.com/facebookresearch/fairseq/tree/main/examples/mms#asr>
        :param log_level: log level
        """
        set_log_level(log_level)
        self.cfg = constants.CFG.copy()
        self.model = Path(model)
        self.cfg['common_eval']['path'] = str(self.model.resolve())

        self.tmp_dir = tempfile.TemporaryDirectory()
        self.tmp_dir_path = Path(self.tmp_dir.name)
        atexit.register(self._cleanup)

        self.wer = None

    def _cleanup(self) -> None:
        """
        cleans up the temp_dir
        :return: None
        """
        self.tmp_dir.cleanup()

    def _setup_tmp_dir(self, media_files: List[str]) -> None:
        """
        Sets up the tmp dir with the required files
        Taken from https://github.com/facebookresearch/fairseq/blob/main/examples/mms/asr/infer/mms_infer.py
        :param media_files: list of media files
        :return: None
        """
        logger.info(f"Setting up tmp dir: {self.tmp_dir}")
        with open(self.tmp_dir_path / "dev.tsv", "w") as fw:
            fw.write("/\n")
            for audio in media_files:
                fw.write(f"{audio[0]}\t{audio[1]}\n")
        with open(self.tmp_dir_path / "dev.uid", "w") as fw:
            fw.write(f"{audio}\n" * len(media_files))
        with open(self.tmp_dir_path / "dev.ltr", "w") as fw:
            fw.write("d u m m y | d u m m y\n" * len(media_files))
        with open(self.tmp_dir_path / "dev.wrd", "w") as fw:
            fw.write("dummy dummy\n" * len(media_files))

    def _prepare_media_files(self, media_files: List[str]) -> List[str]:
        """
        Takes the list of media files, converts them to wav and sets the sample rate to 16k
        :param media_files: list of media files
        :return: list of paths to processed media files
        """
        files = [Path(file) for file in media_files]
        res = []
        for file in files:
            logger.info(f"Preparing file {file}")
            assert file.exists(), FileNotFoundError(file)
            info = mediainfo(file)
            if info['format_name'] == 'wav' and info['sample_rate'] == '16000':
                res.append((str(file.resolve()), info['duration_ts']))
            else:
                seg = AudioSegment.from_file(file, format=info['format_name'])
                processed_file = seg.set_frame_rate(16000)
                ds = processed_file.frame_count()
                save_file_at = self.tmp_dir_path / f"{file.stem}.wav"
                processed_file.export(save_file_at, format='wav')
                res.append((str(save_file_at.resolve()), int(ds)))
        return res

    def transcribe(self,
                   media_files: List[str],
                   lang: str = 'eng',
                   device: str = None,
                   align: bool = False,
                   timestamps_type: str = 'segment',
                   max_segment_len: int = 27,
                   cfg: dict = None) -> Union[List[str], List[dict]]:
        """
        Transcribes a list of media files provided as inputs

        :param media_files: list of media files (video/audio), in whichever format supported by ffmpeg
        :param lang: the language of the media
        :param device: Pytorch device (`cuda`, `cpu` or `tpu`)
        :param align: if True the alignment model will be used to generate the timestamps, otherwise you will get raw text from the MMS model
        :param timestamps_type: Once of (`segment`, `word` or `char`) if `align` is set to True, this will be used to fragment the raw text
        :param max_segment_len: the maximum length of the fragmented segments
        :param cfg: configuration dict in case you want to use a custom configuration, see [CFG](#Constants.CFG)

        :return: List of transcription text in the same order as input files
        """
        processed_files = self._prepare_media_files(media_files)
        cwd = os.getcwd()
        # clone Fairseq
        wow_ai_mms_utils.clone(constants.FAIRSEQ_URL, constants.FAIRSEQ_DIR)
        fairseq_dir = str(constants.FAIRSEQ_DIR.resolve())
        sys.path.append(fairseq_dir)
        os.chdir(fairseq_dir)
        # import
        from wow_ai_mms.models.speech_recognition.new.infer import hydra_main
        try:
            from fairseq.data.data_utils_fast import (
                batch_by_size_fn,
                batch_by_size_vec,
                batch_fixed_shapes_fast,
            )
        except ImportError:
            # we need to build the extension
            logger.info("Bulding required extensions, this may take a while ...")
            from distutils.core import run_setup
            run_setup(str((constants.FAIRSEQ_DIR / 'setup.py').resolve()), script_args=['build_ext', '--inplace'],
                      stop_after='run')


        self._setup_tmp_dir(processed_files)
        # edit cfg
        if cfg is None:
            self.cfg['task']['data'] = self.cfg['decoding']['results_path'] = str(self.tmp_dir_path.resolve())
            self.cfg['dataset']['gen_subset'] = f'{lang}:dev'
            if device is None:
                if torch.cuda.is_available():
                    device = 'cuda'
                else:
                    device = 'cpu'
            if device == 'cuda':
                pass  # default
            elif device == 'cpu':
                self.cfg['common']['cpu'] = True
            if device == 'tpu':
                self.cfg['common']['tpu'] = True
            cfg = OmegaConf.structured(self.cfg)

        self.wer = hydra_main(cfg)
        # get results: will just read from hypo.word as I don't want to change fairseq repo to get the hypo array
        hypo_file = self.tmp_dir_path / constants.HYPO_WORDS_FILE
        res = []
        with open(hypo_file) as hw:
            hypos = hw.readlines()
            outputs = self._reorder_decode(hypos)
            transcripts = [line[1].strip() for line in outputs]
        if align:
            align_model = AlignmentModel()
            for i in range(len(transcripts)):
                media_file = processed_files[i][0]
                transcript = wow_ai_mms_utils.get_transcript_segments(transcripts[i], timestamps_type, max_segment_len=max_segment_len)
                segments = align_model.align(media_file=media_file,
                                             transcript=transcript,
                                             lang=lang,
                                             device=device)
                res.append(segments)
        else:
            res = transcripts

        os.chdir(cwd)
        return res

    @staticmethod
    def get_supported_langs() -> List[str]:
        """
        Helper function to get supported ISO 693-3 languages by the ASR model
        Source <https://dl.fbaipublicfiles.com/mms/misc/language_coverage_mms.html>
        :return: list of supported languages
        """
        with open(constants.MMS_LANGS_FILE) as f:
            data = json.load(f)
            return [key for key in data if data[key]['ASR']]

    @staticmethod
    def _reorder_decode(hypos):
        """
        Helper method to reorder the `hypos`, see @bekarys0504 comment in
        [#5](https://github.com/abdeladim-s/wow_ai_mms/issues/5)

        :param hypos: hypos list
        :return: ordered hypos
        """
        outputs = []
        for hypo in hypos:
            idx = int(re.findall("\(None-(\d+)\)$", hypo)[0])
            hypo = re.sub("\(\S+\)$", "", hypo).strip()
            outputs.append((idx, hypo))
        outputs = sorted(outputs)
        return outputs


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import math

import torch
import torch.nn.functional as F
from fairseq import utils
from fairseq.criterions import FairseqCriterion, register_criterion


@register_criterion("cross_entropy_acc")
class CrossEntropyWithAccCriterion(FairseqCriterion):
    def __init__(self, task, sentence_avg):
        super().__init__(task)
        self.sentence_avg = sentence_avg

    def compute_loss(self, model, net_output, target, reduction, log_probs):
        # N, T -> N * T
        target = target.view(-1)
        lprobs = model.get_normalized_probs(net_output, log_probs=log_probs)
        if not hasattr(lprobs, "batch_first"):
            logging.warning(
                "ERROR: we need to know whether "
                "batch first for the net output; "
                "you need to set batch_first attribute for the return value of "
                "model.get_normalized_probs. Now, we assume this is true, but "
                "in the future, we will raise exception instead. "
            )
        batch_first = getattr(lprobs, "batch_first", True)
        if not batch_first:
            lprobs = lprobs.transpose(0, 1)

        # N, T, D -> N * T, D
        lprobs = lprobs.view(-1, lprobs.size(-1))
        loss = F.nll_loss(
            lprobs, target, ignore_index=self.padding_idx, reduction=reduction
        )
        return lprobs, loss

    def get_logging_output(self, sample, target, lprobs, loss):
        target = target.view(-1)
        mask = target != self.padding_idx
        correct = torch.sum(
            lprobs.argmax(1).masked_select(mask) == target.masked_select(mask)
        )
        total = torch.sum(mask)
        sample_size = (
            sample["target"].size(0) if self.sentence_avg else sample["ntokens"]
        )

        logging_output = {
            "loss": utils.item(loss.data),  # * sample['ntokens'],
            "ntokens": sample["ntokens"],
            "nsentences": sample["target"].size(0),
            "sample_size": sample_size,
            "correct": utils.item(correct.data),
            "total": utils.item(total.data),
            "nframes": torch.sum(sample["net_input"]["src_lengths"]).item(),
        }

        return sample_size, logging_output

    def forward(self, model, sample, reduction="sum", log_probs=True):
        """Computes the cross entropy with accuracy metric for the given sample.

        This is similar to CrossEntropyCriterion in fairseq, but also
        computes accuracy metrics as part of logging

        Args:
            logprobs (Torch.tensor) of shape N, T, D i.e.
                batchsize, timesteps, dimensions
            targets (Torch.tensor) of shape N, T  i.e batchsize, timesteps

        Returns:
        tuple: With three elements:
            1) the loss
            2) the sample size, which is used as the denominator for the gradient
            3) logging outputs to display while training

        TODO:
            * Currently this Criterion will only work with LSTMEncoderModels or
            FairseqModels which have decoder, or Models which return TorchTensor
            as net_output.
            We need to make a change to support all FairseqEncoder models.
        """
        net_output = model(**sample["net_input"])
        target = model.get_targets(sample, net_output)
        lprobs, loss = self.compute_loss(
            model, net_output, target, reduction, log_probs
        )
        sample_size, logging_output = self.get_logging_output(
            sample, target, lprobs, loss
        )
        return loss, sample_size, logging_output

    @staticmethod
    def aggregate_logging_outputs(logging_outputs):
        """Aggregate logging outputs from data parallel training."""
        correct_sum = sum(log.get("correct", 0) for log in logging_outputs)
        total_sum = sum(log.get("total", 0) for log in logging_outputs)
        loss_sum = sum(log.get("loss", 0) for log in logging_outputs)
        ntokens = sum(log.get("ntokens", 0) for log in logging_outputs)
        nsentences = sum(log.get("nsentences", 0) for log in logging_outputs)
        sample_size = sum(log.get("sample_size", 0) for log in logging_outputs)
        nframes = sum(log.get("nframes", 0) for log in logging_outputs)
        agg_output = {
            "loss": loss_sum / sample_size / math.log(2) if sample_size > 0 else 0.0,
            # if args.sentence_avg, then sample_size is nsentences, then loss
            # is per-sentence loss; else sample_size is ntokens, the loss
            # becomes per-output token loss
            "ntokens": ntokens,
            "nsentences": nsentences,
            "nframes": nframes,
            "sample_size": sample_size,
            "acc": correct_sum * 100.0 / total_sum if total_sum > 0 else 0.0,
            "correct": correct_sum,
            "total": total_sum,
            # total is the number of validate tokens
        }
        if sample_size != ntokens:
            agg_output["nll_loss"] = loss_sum / ntokens / math.log(2)
        # loss: per output token loss
        # nll_loss: per sentence loss
        return agg_output



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import re
import glob
import json
import tempfile
import math
import torch
from torch import nn
from torch.nn import functional as F
from torch.utils.data import DataLoader
import numpy as np
import commons
import utils
import argparse
import subprocess
from data_utils import TextAudioLoader, TextAudioCollate, TextAudioSpeakerLoader, TextAudioSpeakerCollate
from models import SynthesizerTrn
from scipy.io.wavfile import write

class TextMapper(object):
    def __init__(self, vocab_file):
        self.symbols = [x.replace("\n", "") for x in open(vocab_file, encoding="utf-8").readlines()]
        self.SPACE_ID = self.symbols.index(" ")
        self._symbol_to_id = {s: i for i, s in enumerate(self.symbols)}
        self._id_to_symbol = {i: s for i, s in enumerate(self.symbols)}

    def text_to_sequence(self, text, cleaner_names):
        '''Converts a string of text to a sequence of IDs corresponding to the symbols in the text.
        Args:
        text: string to convert to a sequence
        cleaner_names: names of the cleaner functions to run the text through
        Returns:
        List of integers corresponding to the symbols in the text
        '''
        sequence = []
        clean_text = text.strip()
        for symbol in clean_text:
            symbol_id = self._symbol_to_id[symbol]
            sequence += [symbol_id]
        return sequence

    def uromanize(self, text, uroman_pl):
        iso = "xxx"
        with tempfile.NamedTemporaryFile() as tf, \
             tempfile.NamedTemporaryFile() as tf2:
            with open(tf.name, "w") as f:
                f.write("\n".join([text]))
            cmd = f"perl " + uroman_pl
            cmd += f" -l {iso} "
            cmd +=  f" < {tf.name} > {tf2.name}"
            os.system(cmd)
            outtexts = []
            with open(tf2.name) as f:
                for line in f:
                    line =  re.sub(r"\s+", " ", line).strip()
                    outtexts.append(line)
            outtext = outtexts[0]
        return outtext

    def get_text(self, text, hps):
        text_norm = self.text_to_sequence(text, hps.data.text_cleaners)
        if hps.data.add_blank:
            text_norm = commons.intersperse(text_norm, 0)
        text_norm = torch.LongTensor(text_norm)
        return text_norm

    def filter_oov(self, text, lang=None):
        text = self.preprocess_char(text, lang=lang)
        val_chars = self._symbol_to_id
        txt_filt = "".join(list(filter(lambda x: x in val_chars, text)))
        print(f"text after filtering OOV: {txt_filt}")
        return txt_filt

    def preprocess_char(self, text, lang=None):
        """
        Special treatement of characters in certain languages
        """
        if lang == "ron":
            text = text.replace("ț", "ţ")
            print(f"{lang} (ț -> ţ): {text}")
        return text

def generate():
    parser = argparse.ArgumentParser(description='TTS inference')
    parser.add_argument('--model-dir', type=str, help='model checkpoint dir')
    parser.add_argument('--wav', type=str, help='output wav path')
    parser.add_argument('--txt', type=str, help='input text')
    parser.add_argument('--uroman-dir', type=str, default=None, help='uroman lib dir (will download if not specified)')
    parser.add_argument('--lang', type=str, default=None, help='language iso code (required for Romanian)')
    args = parser.parse_args()
    ckpt_dir, wav_path, txt = args.model_dir, args.wav, args.txt

    if torch.cuda.is_available():
        device = torch.device("cuda")
    elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available() and torch.backends.mps.is_built():
        device = torch.device("mps")
    else:
        device = torch.device("cpu")

    print(f"Run inference with {device}")
    vocab_file = f"{ckpt_dir}/vocab.txt"
    config_file = f"{ckpt_dir}/config.json"
    assert os.path.isfile(config_file), f"{config_file} doesn't exist"
    hps = utils.get_hparams_from_file(config_file)
    text_mapper = TextMapper(vocab_file)
    net_g = SynthesizerTrn(
        len(text_mapper.symbols),
        hps.data.filter_length // 2 + 1,
        hps.train.segment_size // hps.data.hop_length,
        **hps.model)
    net_g.to(device)
    _ = net_g.eval()

    g_pth = f"{ckpt_dir}/G_100000.pth"
    print(f"load {g_pth}")

    _ = utils.load_checkpoint(g_pth, net_g, None)

    print(f"text: {txt}")
    is_uroman = hps.data.training_files.split('.')[-1] == 'uroman'
    if is_uroman:
        with tempfile.TemporaryDirectory() as tmp_dir:
            if args.uroman_dir is None:
                cmd = f"git clone git@github.com:isi-nlp/uroman.git {tmp_dir}"
                print(cmd)
                subprocess.check_output(cmd, shell=True)
                args.uroman_dir = tmp_dir
            uroman_pl = os.path.join(args.uroman_dir, "bin", "uroman.pl")
            print(f"uromanize")
            txt = text_mapper.uromanize(txt, uroman_pl)
            print(f"uroman text: {txt}")
    txt = txt.lower()
    txt = text_mapper.filter_oov(txt, lang=args.lang)
    stn_tst = text_mapper.get_text(txt, hps)
    with torch.no_grad():
        x_tst = stn_tst.unsqueeze(0).to(device)
        x_tst_lengths = torch.LongTensor([stn_tst.size(0)]).to(device)
        hyp = net_g.infer(
            x_tst, x_tst_lengths, noise_scale=.667,
            noise_scale_w=0.8, length_scale=1.0
        )[0][0,0].cpu().float().numpy()

    os.makedirs(os.path.dirname(wav_path), exist_ok=True)
    print(f"wav: {wav_path}")
    write(wav_path, hps.data.sampling_rate, hyp)
    return


if __name__ == '__main__':
    generate()



import importlib
import os


for file in sorted(os.listdir(os.path.dirname(__file__))):
    if file.endswith(".py") and not file.startswith("_"):
        task_name = file[: file.find(".py")]
        importlib.import_module("wow_ai_mms.models.speech_recognition.tasks." + task_name)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from concurrent.futures import ThreadPoolExecutor
import logging
from omegaconf import MISSING
import os
import torch
from typing import Optional
import warnings


from dataclasses import dataclass
from fairseq.dataclass import FairseqDataclass
from .kaldi_initializer import KaldiInitializerConfig, initalize_kaldi


logger = logging.getLogger(__name__)


@dataclass
class KaldiDecoderConfig(FairseqDataclass):
    hlg_graph_path: Optional[str] = None
    output_dict: str = MISSING

    kaldi_initializer_config: Optional[KaldiInitializerConfig] = None

    acoustic_scale: float = 0.5
    max_active: int = 10000
    beam_delta: float = 0.5
    hash_ratio: float = 2.0

    is_lattice: bool = False
    lattice_beam: float = 10.0
    prune_interval: int = 25
    determinize_lattice: bool = True
    prune_scale: float = 0.1
    max_mem: int = 0
    phone_determinize: bool = True
    word_determinize: bool = True
    minimize: bool = True

    num_threads: int = 1


class KaldiDecoder(object):
    def __init__(
        self,
        cfg: KaldiDecoderConfig,
        beam: int,
        nbest: int = 1,
    ):
        try:
            from kaldi.asr import FasterRecognizer, LatticeFasterRecognizer
            from kaldi.base import set_verbose_level
            from kaldi.decoder import (
                FasterDecoder,
                FasterDecoderOptions,
                LatticeFasterDecoder,
                LatticeFasterDecoderOptions,
            )
            from kaldi.lat.functions import DeterminizeLatticePhonePrunedOptions
            from kaldi.fstext import read_fst_kaldi, SymbolTable
        except:
            warnings.warn(
                "pykaldi is required for this functionality. Please install from https://github.com/pykaldi/pykaldi"
            )

        # set_verbose_level(2)

        self.acoustic_scale = cfg.acoustic_scale
        self.nbest = nbest

        if cfg.hlg_graph_path is None:
            assert (
                cfg.kaldi_initializer_config is not None
            ), "Must provide hlg graph path or kaldi initializer config"
            cfg.hlg_graph_path = initalize_kaldi(cfg.kaldi_initializer_config)

        assert os.path.exists(cfg.hlg_graph_path), cfg.hlg_graph_path

        if cfg.is_lattice:
            self.dec_cls = LatticeFasterDecoder
            opt_cls = LatticeFasterDecoderOptions
            self.rec_cls = LatticeFasterRecognizer
        else:
            assert self.nbest == 1, "nbest > 1 requires lattice decoder"
            self.dec_cls = FasterDecoder
            opt_cls = FasterDecoderOptions
            self.rec_cls = FasterRecognizer

        self.decoder_options = opt_cls()
        self.decoder_options.beam = beam
        self.decoder_options.max_active = cfg.max_active
        self.decoder_options.beam_delta = cfg.beam_delta
        self.decoder_options.hash_ratio = cfg.hash_ratio

        if cfg.is_lattice:
            self.decoder_options.lattice_beam = cfg.lattice_beam
            self.decoder_options.prune_interval = cfg.prune_interval
            self.decoder_options.determinize_lattice = cfg.determinize_lattice
            self.decoder_options.prune_scale = cfg.prune_scale
            det_opts = DeterminizeLatticePhonePrunedOptions()
            det_opts.max_mem = cfg.max_mem
            det_opts.phone_determinize = cfg.phone_determinize
            det_opts.word_determinize = cfg.word_determinize
            det_opts.minimize = cfg.minimize
            self.decoder_options.det_opts = det_opts

        self.output_symbols = {}
        with open(cfg.output_dict, "r") as f:
            for line in f:
                items = line.rstrip().split()
                assert len(items) == 2
                self.output_symbols[int(items[1])] = items[0]

        logger.info(f"Loading FST from {cfg.hlg_graph_path}")
        self.fst = read_fst_kaldi(cfg.hlg_graph_path)
        self.symbol_table = SymbolTable.read_text(cfg.output_dict)

        self.executor = ThreadPoolExecutor(max_workers=cfg.num_threads)

    def generate(self, models, sample, **unused):
        """Generate a batch of inferences."""
        # model.forward normally channels prev_output_tokens into the decoder
        # separately, but SequenceGenerator directly calls model.encoder
        encoder_input = {
            k: v for k, v in sample["net_input"].items() if k != "prev_output_tokens"
        }
        emissions, padding = self.get_emissions(models, encoder_input)
        return self.decode(emissions, padding)

    def get_emissions(self, models, encoder_input):
        """Run encoder and normalize emissions"""
        model = models[0]

        all_encoder_out = [m(**encoder_input) for m in models]

        if len(all_encoder_out) > 1:

            if "encoder_out" in all_encoder_out[0]:
                encoder_out = {
                    "encoder_out": sum(e["encoder_out"] for e in all_encoder_out)
                    / len(all_encoder_out),
                    "encoder_padding_mask": all_encoder_out[0]["encoder_padding_mask"],
                }
                padding = encoder_out["encoder_padding_mask"]
            else:
                encoder_out = {
                    "logits": sum(e["logits"] for e in all_encoder_out)
                    / len(all_encoder_out),
                    "padding_mask": all_encoder_out[0]["padding_mask"],
                }
                padding = encoder_out["padding_mask"]
        else:
            encoder_out = all_encoder_out[0]
            padding = (
                encoder_out["padding_mask"]
                if "padding_mask" in encoder_out
                else encoder_out["encoder_padding_mask"]
            )

        if hasattr(model, "get_logits"):
            emissions = model.get_logits(encoder_out, normalize=True)
        else:
            emissions = model.get_normalized_probs(encoder_out, log_probs=True)

        return (
            emissions.cpu().float().transpose(0, 1),
            padding.cpu() if padding is not None and padding.any() else None,
        )

    def decode_one(self, logits, padding):
        from kaldi.matrix import Matrix

        decoder = self.dec_cls(self.fst, self.decoder_options)
        asr = self.rec_cls(
            decoder, self.symbol_table, acoustic_scale=self.acoustic_scale
        )

        if padding is not None:
            logits = logits[~padding]

        mat = Matrix(logits.numpy())

        out = asr.decode(mat)

        if self.nbest > 1:
            from kaldi.fstext import shortestpath
            from kaldi.fstext.utils import (
                convert_compact_lattice_to_lattice,
                convert_lattice_to_std,
                convert_nbest_to_list,
                get_linear_symbol_sequence,
            )

            lat = out["lattice"]

            sp = shortestpath(lat, nshortest=self.nbest)

            sp = convert_compact_lattice_to_lattice(sp)
            sp = convert_lattice_to_std(sp)
            seq = convert_nbest_to_list(sp)

            results = []
            for s in seq:
                _, o, w = get_linear_symbol_sequence(s)
                words = list(self.output_symbols[z] for z in o)
                results.append(
                    {
                        "tokens": words,
                        "words": words,
                        "score": w.value,
                        "emissions": logits,
                    }
                )
            return results
        else:
            words = out["text"].split()
            return [
                {
                    "tokens": words,
                    "words": words,
                    "score": out["likelihood"],
                    "emissions": logits,
                }
            ]

    def decode(self, emissions, padding):
        if padding is None:
            padding = [None] * len(emissions)

        ret = list(
            map(
                lambda e, p: self.executor.submit(self.decode_one, e, p),
                emissions,
                padding,
            )
        )
        return ret



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .asr_dataset import AsrDataset


__all__ = [
    "AsrDataset",
]



import json
import re
import unicodedata

from examples.mms.data_prep.norm_config import norm_config


def text_normalize(text, iso_code, lower_case=True, remove_numbers=True, remove_brackets=False):

    """Given a text, normalize it by changing to lower case, removing punctuations, removing words that only contain digits and removing extra spaces

    Args:
        text : The string to be normalized
        iso_code :
        remove_numbers : Boolean flag to specify if words containing only digits should be removed

    Returns:
        normalized_text : the string after all normalization  

    """

    config = norm_config.get(iso_code, norm_config["*"])

    for field in ["lower_case", "punc_set","del_set", "mapping", "digit_set", "unicode_norm"]:
        if field not in config:
            config[field] = norm_config["*"][field]


    text = unicodedata.normalize(config["unicode_norm"], text)

    # Convert to lower case

    if config["lower_case"] and lower_case:
        text = text.lower()

    # brackets
    
    # always text inside brackets with numbers in them. Usually corresponds to "(Sam 23:17)"
    text = re.sub(r"\([^\)]*\d[^\)]*\)", " ", text)
    if remove_brackets:
        text = re.sub(r"\([^\)]*\)", " ", text)

    # Apply mappings

    for old, new in config["mapping"].items():
        text = re.sub(old, new, text)

    # Replace punctutations with space

    punct_pattern = r"[" + config["punc_set"]

    punct_pattern += "]"

    normalized_text = re.sub(punct_pattern, " ", text)

    # remove characters in delete list

    delete_patten = r"[" + config["del_set"] + "]"

    normalized_text = re.sub(delete_patten, "", normalized_text)

    # Remove words containing only digits
    # We check for 3 cases  a)text starts with a number b) a number is present somewhere in the middle of the text c) the text ends with a number
    # For each case we use lookaround regex pattern to see if the digit pattern in preceded and followed by whitespaces, only then we replace the numbers with space
    # The lookaround enables overlapping pattern matches to be replaced

    if remove_numbers:

        digits_pattern = "[" + config["digit_set"]

        digits_pattern += "]+"

        complete_digit_pattern = (
            r"^"
            + digits_pattern
            + "(?=\s)|(?<=\s)"
            + digits_pattern
            + "(?=\s)|(?<=\s)"
            + digits_pattern
            + "$"
        )

        normalized_text = re.sub(complete_digit_pattern, " ", normalized_text)

    if config["rm_diacritics"]:
        from unidecode import unidecode
        normalized_text = unidecode(normalized_text)

    # Remove extra spaces
    normalized_text = re.sub(r"\s+", " ", normalized_text).strip()

    return normalized_text



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
    This module contains collection of classes which implement
    collate functionalities for various tasks.

    Collaters should know what data to expect for each sample
    and they should pack / collate them into batches
"""


from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
import torch
from fairseq.data import data_utils as fairseq_data_utils


class Seq2SeqCollater(object):
    """
    Implements collate function mainly for seq2seq tasks
    This expects each sample to contain feature (src_tokens) and
    targets.
    This collator is also used for aligned training task.
    """

    def __init__(
        self,
        feature_index=0,
        label_index=1,
        pad_index=1,
        eos_index=2,
        move_eos_to_beginning=True,
    ):
        self.feature_index = feature_index
        self.label_index = label_index
        self.pad_index = pad_index
        self.eos_index = eos_index
        self.move_eos_to_beginning = move_eos_to_beginning

    def _collate_frames(self, frames):
        """Convert a list of 2d frames into a padded 3d tensor
        Args:
            frames (list): list of 2d frames of size L[i]*f_dim. Where L[i] is
                length of i-th frame and f_dim is static dimension of features
        Returns:
            3d tensor of size len(frames)*len_max*f_dim where len_max is max of L[i]
        """
        len_max = max(frame.size(0) for frame in frames)
        f_dim = frames[0].size(1)
        res = frames[0].new(len(frames), len_max, f_dim).fill_(0.0)

        for i, v in enumerate(frames):
            res[i, : v.size(0)] = v

        return res

    def collate(self, samples):
        """
        utility function to collate samples into batch for speech recognition.
        """
        if len(samples) == 0:
            return {}

        # parse samples into torch tensors
        parsed_samples = []
        for s in samples:
            # skip invalid samples
            if s["data"][self.feature_index] is None:
                continue
            source = s["data"][self.feature_index]
            if isinstance(source, (np.ndarray, np.generic)):
                source = torch.from_numpy(source)
            target = s["data"][self.label_index]
            if isinstance(target, (np.ndarray, np.generic)):
                target = torch.from_numpy(target).long()
            elif isinstance(target, list):
                target = torch.LongTensor(target)

            parsed_sample = {"id": s["id"], "source": source, "target": target}
            parsed_samples.append(parsed_sample)
        samples = parsed_samples

        id = torch.LongTensor([s["id"] for s in samples])
        frames = self._collate_frames([s["source"] for s in samples])
        # sort samples by descending number of frames
        frames_lengths = torch.LongTensor([s["source"].size(0) for s in samples])
        frames_lengths, sort_order = frames_lengths.sort(descending=True)
        id = id.index_select(0, sort_order)
        frames = frames.index_select(0, sort_order)

        target = None
        target_lengths = None
        prev_output_tokens = None
        if samples[0].get("target", None) is not None:
            ntokens = sum(len(s["target"]) for s in samples)
            target = fairseq_data_utils.collate_tokens(
                [s["target"] for s in samples],
                self.pad_index,
                self.eos_index,
                left_pad=False,
                move_eos_to_beginning=False,
            )
            target = target.index_select(0, sort_order)
            target_lengths = torch.LongTensor(
                [s["target"].size(0) for s in samples]
            ).index_select(0, sort_order)
            prev_output_tokens = fairseq_data_utils.collate_tokens(
                [s["target"] for s in samples],
                self.pad_index,
                self.eos_index,
                left_pad=False,
                move_eos_to_beginning=self.move_eos_to_beginning,
            )
            prev_output_tokens = prev_output_tokens.index_select(0, sort_order)
        else:
            ntokens = sum(len(s["source"]) for s in samples)

        batch = {
            "id": id,
            "ntokens": ntokens,
            "net_input": {"src_tokens": frames, "src_lengths": frames_lengths},
            "target": target,
            "target_lengths": target_lengths,
            "nsentences": len(samples),
        }
        if prev_output_tokens is not None:
            batch["net_input"]["prev_output_tokens"] = prev_output_tokens
        return batch



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
from dataclasses import dataclass, field
from typing import Optional

from fairseq.dataclass.configs import FairseqDataclass
from fairseq.dataclass.constants import ChoiceEnum
from omegaconf import MISSING


DECODER_CHOICES = ChoiceEnum(["viterbi", "kenlm", "fairseqlm"])


@dataclass
class DecoderConfig(FairseqDataclass):
    type: DECODER_CHOICES = field(
        default="viterbi",
        metadata={"help": "The type of decoder to use"},
    )


@dataclass
class FlashlightDecoderConfig(FairseqDataclass):
    nbest: int = field(
        default=1,
        metadata={"help": "Number of decodings to return"},
    )
    unitlm: bool = field(
        default=False,
        metadata={"help": "If set, use unit language model"},
    )
    lmpath: str = field(
        default=MISSING,
        metadata={"help": "Language model for KenLM decoder"},
    )
    lexicon: Optional[str] = field(
        default=None,
        metadata={"help": "Lexicon for Flashlight decoder"},
    )
    beam: int = field(
        default=50,
        metadata={"help": "Number of beams to use for decoding"},
    )
    beamthreshold: float = field(
        default=50.0,
        metadata={"help": "Threshold for beam search decoding"},
    )
    beamsizetoken: Optional[int] = field(
        default=None, metadata={"help": "Beam size to use"}
    )
    wordscore: float = field(
        default=-1,
        metadata={"help": "Word score for KenLM decoder"},
    )
    unkweight: float = field(
        default=-math.inf,
        metadata={"help": "Unknown weight for KenLM decoder"},
    )
    silweight: float = field(
        default=0,
        metadata={"help": "Silence weight for KenLM decoder"},
    )
    lmweight: float = field(
        default=2,
        metadata={"help": "Weight for LM while interpolating score"},
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file contains a simple class to use the AlignmentModel model from the MMS project
More info can be found here: https://github.com/facebookresearch/fairseq/tree/main/examples/mms/data_prep
"""
import logging
import shutil
import sys
from pathlib import Path
from typing import List
import torch
from wow_ai_mms import constants
# fix importing from fairseq.examples
# import site
# sys.path.append(str(Path(site.getsitepackages()[0]) / 'fairseq'))
# try:
#     from fairseq.examples.mms.data_prep.align_and_segment import get_alignments
#     from fairseq.examples.mms.data_prep.align_utils import get_uroman_tokens, get_spans
#     from fairseq.examples.mms.data_prep.text_normalization import text_normalize
# except ImportError:
#     from examples.mms.data_prep.align_and_segment import get_alignments
#     from examples.mms.data_prep.align_utils import get_uroman_tokens, get_spans
#     from examples.mms.data_prep.text_normalization import text_normalize

from wow_ai_mms import utils
from wow_ai_mms._logger import set_log_level
from wow_ai_mms.constants import PACKAGE_DATA_DIR, ALIGNMENT_MODEL_URL, ALIGNMENT_DICTIONARY_URL, UROMAN_URL, \
    UROMAN_DIR



from torchaudio.models import wav2vec2_model

logger = logging.getLogger(__name__)


class AlignmentModel:
    """
    MMS Alignment algorithm
    Example usage:

    ```python
    from wow_ai_mms.models.alignment import AlignmentModel

    align_model = AlignmentModel()
    transcriptions = align_model.align('/home/su/code/wow_ai_mms/assets/eng_1.mp3',
                                       transcript=["segment 1", "segment 2"],
                                       lang='eng')
    for transcription in transcriptions:
        for segment in transcription:
            print(f"{segment['start_time']} -> {segment['end_time']}: {segment['text']}")
    ```
    """
    def __init__(self,
                 model: str = None,
                 dictionary: str = None,
                 uroman_dir: str = None,
                 log_level: int = logging.INFO):

        set_log_level(log_level)
        assert shutil.which("perl") is not None, "To use the alignment algorithm you will need uroman " \
                                                 "<https://github.com/isi-nlp/uroman> which is written in perl " \
                                                 "please install perl first <https://www.perl.org/get.html>"
        if uroman_dir is not None:
            self.uroman_dir_path = Path(uroman_dir)
        else:
            self.uroman_dir_path = utils.get_uroman()

        if model is not None:
            self.model_path = Path(model)
        else:
            self.model_path = Path(PACKAGE_DATA_DIR) / "ctc_alignment_mling_uroman_model.pt"

        if dictionary is not None:
            self.dictionary_path = Path(dictionary)
        else:
            self.dictionary_path = Path(PACKAGE_DATA_DIR) / "ctc_alignment_mling_uroman_model.dict"

        self.model, self.dictionary = self._load_model_dict()

        # clone Fairseq
        utils.clone(constants.FAIRSEQ_URL, constants.FAIRSEQ_DIR)
        sys.path.append(str(constants.FAIRSEQ_DIR.resolve()))

    def _load_model_dict(self):
        """
        Modified from <https://github.com/facebookresearch/fairseq/blob/main/examples/mms/data_prep/align_utils.py>
        to store the models in a consistent directory

        :return: None
        """
        # model_path_name = self.model
        logger.info(f"Loading AlignmentModel model {str(self.model_path.resolve())}")
        if self.model_path.exists():
            logger.info(f"Using model {self.model_path}")
        else:
            logging.info("Downloading alignment model ...")
            torch.hub.download_url_to_file(
                ALIGNMENT_MODEL_URL,
                str(self.model_path.resolve()),
            )
            assert self.model_path.exists()
        state_dict = torch.load(self.model_path, map_location="cpu")

        model = wav2vec2_model(
            extractor_mode="layer_norm",
            extractor_conv_layer_config=[
                (512, 10, 5),
                (512, 3, 2),
                (512, 3, 2),
                (512, 3, 2),
                (512, 3, 2),
                (512, 2, 2),
                (512, 2, 2),
            ],
            extractor_conv_bias=True,
            encoder_embed_dim=1024,
            encoder_projection_dropout=0.0,
            encoder_pos_conv_kernel=128,
            encoder_pos_conv_groups=16,
            encoder_num_layers=24,
            encoder_num_heads=16,
            encoder_attention_dropout=0.0,
            encoder_ff_interm_features=4096,
            encoder_ff_interm_dropout=0.1,
            encoder_dropout=0.0,
            encoder_layer_norm_first=True,
            encoder_layer_drop=0.1,
            aux_num_out=31,
        )
        model.load_state_dict(state_dict)
        model.eval()

        if self.dictionary_path.exists():
            logger.info("Dictionary path already exists.")
        else:
            logger.info("Downloading dictionary ...")
            torch.hub.download_url_to_file(
                ALIGNMENT_DICTIONARY_URL,
                str(self.dictionary_path.resolve()),
            )
            assert self.dictionary_path.exists()
        with open(self.dictionary_path) as f:
            dictionary = {l.strip(): i for i, l in enumerate(f.readlines())}
        return model, dictionary

    def align(self,
              media_file: str,
              transcript: List[str],
              lang: str,
              device: str = None) -> List[dict]:
        """
        Takes a media file, transcription segments and the lang and returns a list of dicts in the following format
        [{
            'start_time': ...
            'end_time': ...,
            'text': ...,
            'duration': ...
        }, ...]

        :param media_file: the path of the media file, should be wav
        :param transcript: list of segments
        :param lang: language ISO code
        :param device: 'cuda' or 'cpu'
        :return: list of transcription timestamps
        """
        # import
        import os
        cwd = os.getcwd()
        os.chdir(constants.FAIRSEQ_DIR)
        from examples.mms.data_prep.align_and_segment import get_alignments
        from examples.mms.data_prep.align_utils import get_uroman_tokens, get_spans
        from examples.mms.data_prep.text_normalization import text_normalize
        os.chdir(cwd)

        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        model = self.model.to(device)
        res = []
        logger.info(f"Aligning file {media_file} ...")
        norm_transcripts = [text_normalize(line.strip(), lang) for line in transcript]
        tokens = get_uroman_tokens(norm_transcripts, str(self.uroman_dir_path.resolve()), lang)

        segments, stride = get_alignments(
            media_file,
            tokens,
            model,
            self.dictionary,
            use_star=False,
        )
        # Get spans of each line in input text file
        spans = get_spans(tokens, segments)

        for i, t in enumerate(transcript):
            span = spans[i]
            seg_start_idx = span[0].start
            seg_end_idx = span[-1].end
            audio_start_sec = seg_start_idx * stride / 1000
            audio_end_sec = seg_end_idx * stride / 1000
            res.append({'start_time': audio_start_sec,
                        'end_time': audio_end_sec,
                        'text': t,
                        'duration': audio_end_sec - audio_start_sec})

        return res
